#include <GL/gl.h>
#include <Inventor/SbBox.h>
#include <Inventor/SoPrimitiveVertex.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/misc/SoState.h>
#include <Inventor/SbLinear.h>

#include <iostream.h>
#include <math.h>

#include "EmissiveSampleSet.h"

// Inventor macro for defining common node functions
SO_NODE_SOURCE(EmissiveSampleSet);

//////////////////////////////////////////////////////////////////////////////
//
//  Description:
//    This is the function which must be called before attempting to
//    instantiate an object of this type.  This is typically done following
//    a call to either SoDB::init() or SoXt::init() like so
//
//    ....
//    EmissiveSampleSet::initClass();
//    ....
//
//////////////////////////////////////////////////////////////////////////////
void
EmissiveSampleSet::initClass ()
{
  SO_NODE_INIT_CLASS(EmissiveSampleSet, SoShape, "Shape");
}

//////////////////////////////////////////////////////////////////////////////
//
//  Description:
//    All fields for the node are created here by calling the Inventor macro
//    SO_NODE_ADD_FIELD().  The field name and default value are to be
//    supplied here.  The default value must be a valid setValue() function
//    of the field's class.  For example, if a field of type SoMFVec3f is
//    being created the default could be set by
//    myVec3f.setValue(0.0, 0.0, 0.0).
//
//    Notice that the default values have parenthesis around them.  Since
//    this is a macro the text supplied is simply appended to 'setValue'.
//
//////////////////////////////////////////////////////////////////////////////
EmissiveSampleSet::EmissiveSampleSet ()
{
  SO_NODE_CONSTRUCTOR(EmissiveSampleSet);

  SO_NODE_ADD_FIELD(location, (0.0, 0.0, 0.0));
  SO_NODE_ADD_FIELD(normal,   (0.0, 0.0, 1.0));
  SO_NODE_ADD_FIELD(color,    (0.0, 0.0, 0.0));
  SO_NODE_ADD_FIELD(radius,             (0.1));
  SO_NODE_ADD_FIELD(singleRadius,       (0.1));

  SO_NODE_ADD_FIELD(touch,             (TRUE));

  location.setNum(0);
}

//////////////////////////////////////////////////////////////////////////////
//
//  Description:
//    Nothing to delete
//
//////////////////////////////////////////////////////////////////////////////
EmissiveSampleSet::~EmissiveSampleSet () {}

//////////////////////////////////////////////////////////////////////////////
//
//  Description:
//    All the important aspects happen here.  Everytime Inventor redraws the
//    scene graph it calls this function for each SoShape node.  For this
//    class we want to draw n number of samples as efficiently as possible.
//
//////////////////////////////////////////////////////////////////////////////
void
EmissiveSampleSet::GLRender (SoGLRenderAction *action)
{
  // grab the current GL state
  SoState *state = action->getState();

  // make sure we need to render
  if (!shouldGLRender(action))
    return;

  // we want to call as few 'getXXXXX()' functions from the fields as
  // possible.  Each time one is called it triggers Inventor to check the
  // fields connections.  If the field is connected to an Engine then this
  // could potentially slow rendering down if too many 'getXXXXX()' calls are
  // made
  int numSamples        = location.getNum();
  const SbVec3f *coords = location.getValues(0);
  const SbVec3f *norms  = normal.getValues(0);
  const SbColor *col    = color.getValues(0);
  //const float   *rad    = radius.getValues(0);
  const float rad = singleRadius.getValue();

  // set up some variables for rotating the disc
  SbRotation rotate;
  SbVec3f axis, trueNormal = SbVec3f(0.0, 0.0, 1.0);
  float radians, radToDeg = 180.0 / M_PI;

  // Time to make some GL calls
  GLfloat black[4] = {0.0, 0.0, 0.0, 1.0};

  // Since we are playing with the Material properties we need to make sure
  // the previous state is restored, otherwise the rest of the scene graph
  // will look funny.  glPushMatrix() and glPopMatrix() don't restore
  // Material properties.  I think?  All I know is if you take these lines of
  // code out the rest of the scene graph gets screwed up.
  GLfloat oldDiffuse[4], oldAmbient[4], oldEmissive[4];
  glGetMaterialfv(GL_FRONT, GL_DIFFUSE, oldDiffuse);
  glGetMaterialfv(GL_FRONT, GL_AMBIENT, oldAmbient);
  glGetMaterialfv(GL_FRONT, GL_EMISSION, oldEmissive);

  glPushMatrix();
  
  //glEnable(GL_COLOR_MATERIAL);
  
  // set the diffuse and ambient components to black so that only emissive is
  // being colored
  glMaterialfv(GL_FRONT, GL_DIFFUSE, black);
  glMaterialfv(GL_FRONT, GL_AMBIENT, black);
  // set GL to interpret all glColor* calls as emissive colors
  //glColorMaterial(GL_FRONT, GL_EMISSION);

  // compile a list of GL commands for rendering.  Creating a list here will
  // speed up rendering times when rendering several thousand samples in one
  // pass.
  GLuint polyName = glGenLists(1);
  glNewList(polyName, GL_COMPILE);
    glBegin(GL_TRIANGLE_FAN);
      glVertex3f( 0.0,  0.0,    0.15);
      glVertex3f( 1.0,  0.0,   -0.15);
      glVertex3f( 0.5,  0.866, -0.15);
      glVertex3f(-0.5,  0.866, -0.15);
      glVertex3f(-1.0,  0.0,   -0.15);
      glVertex3f(-0.5, -0.866, -0.15);
      glVertex3f( 0.5, -0.866, -0.15);
      glVertex3f( 1.0,  0.0,   -0.15);
    glEnd();
  glEndList();

  // Each sample will have the same normal, but the final shape will be
  // rotated according to the 'normal' field
  glNormal3fv(trueNormal.getValue());
  for (int i = 0; i < numSamples; i++) {
    // determine the axis and angle for the rotation of the current sample
    rotate.setValue(trueNormal, norms[i].getValue());
    rotate.getValue(axis, radians);

    glPushMatrix();
      // Transformation
      glTranslatef(coords[i][0], coords[i][1], coords[i][2]);
      glRotatef(radians * radToDeg, axis[0], axis[1], axis[2]);
      //glScalef(rad[i], rad[i], rad[i]);
      glScalef(rad, rad, rad);

      // Color
      glColor3f(col[i][0], col[i][1], col[i][2]);

      // Draw the disc
      glCallList(polyName);

    glPopMatrix();
  }

  //glDisable(GL_COLOR_MATERIAL);

  // set material properties to original
  //glColorMaterial(GL_FRONT, GL_DIFFUSE);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, oldDiffuse);
  glMaterialfv(GL_FRONT, GL_AMBIENT, oldAmbient);
  glMaterialfv(GL_FRONT, GL_EMISSION, oldEmissive);
  //glColor3fv(oldDiffuse);
  
  glPopMatrix();
}

//////////////////////////////////////////////////////////////////////////////
//
//  Description:
//    This should be implemented for compatibility issues.
//
//////////////////////////////////////////////////////////////////////////////
void
EmissiveSampleSet::generatePrimitives (SoAction *action)
{
  ;
}

//////////////////////////////////////////////////////////////////////////////
//
//  Description:
//    Compute the bounding box of all the sample combined.  This is the
//    minimum and maximum point out of all of them.
//
//////////////////////////////////////////////////////////////////////////////
void
EmissiveSampleSet::computeBBox (SoAction *action, SbBox3f &box, SbVec3f &ctr)
{
  SbVec3f min, max;
  int i, j, num = location.getNum();
  const SbVec3f *pos = location.getValues(0);
  const float *rad = radius.getValues(0);
  float maxRad;
  SbVec3f offset;

  if (num < 1) {
    offset = SbVec3f(0.1, 0.1, 0.1);
    SbVec3f ctr = SbVec3f(0.0, 0.0, 0.0);
    box.setBounds(ctr - offset, ctr + offset);
    return;
  }

  maxRad = singleRadius.getValue(); //rad[0];
  min    = pos[0];
  max    = pos[0];

  for (i = 1; i < num; i++) {
    for (j = 0; j < 3; j++) {
      if (pos[i][j] < min[j])
	min[j] = pos[i][j];
      else if (pos[i][j] > max[j])
	max[j] = pos[i][j];
    }
    
    //if (maxRad < rad[i])
    //  maxRad = rad[i];
  }

  offset = SbVec3f(maxRad, maxRad, maxRad);
  box.setBounds(min - offset, max + offset);
  ctr = (min + max) / 2.0;
}

  

  
  

