// Kevin Beason
// http://www.cs.fsu.edu/~beason/
// 4/16/02
//
// This Open Inventor application loads a data set, in this case a reduced
// version of the stanford bunny, and sends rays through the volume. It
// interprets the scalar data as the density and integrates it, producing white
// for high total density and black for low total density. (Total density is
// just the integration/sum of the density along the ray).
// 
// The program keeps running, continually producing an ever finer display of the
// volume. However, over time, the results do not seem to continue to get
// better and I am not sure why. Maybe I am not scaling the radius the samples
// correctly.
// 
// Anyhow, this program is hand tweaked to load the reduced bunny data set.
// Instead of including the bunny data set, I provide a link to the original
// source and the command I used to reduce it.
// 
// Bunny data:
// http://graphics.stanford.edu/data/voldata/
// 
// Reduction program:
// ../hw06/deci.cxx
// 
// Command:
// deci -i bunny 512 512 362 -n 2 -o bunny2 8

#include <math.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <string.h>
#include <stdio.h>

#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoCone.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/sensors/SoTimerSensor.h>
#include <Inventor/nodes/SoComplexity.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoDB.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/SbLinear.h>
#include <Inventor/nodes/SoShapeHints.h>
#include <Inventor/draggers/SoTranslate1Dragger.h>
#include <Inventor/engines/SoCalculator.h>
#include <Inventor/nodes/SoVertexProperty.h>
#include <Inventor/nodes/SoIndexedFaceSet.h>
#include <Inventor/fields/SoSubField.h>
#include <Inventor/SbLinear.h>
#include <Inventor/nodes/SoIndexedLineSet.h>

#include "EmissiveSampleSet.cpp"

int numSamples = 0;
int rendSamplesMax = 100000;
int rendSamplesPerPass = 15;
int rendSamples = 0;
double myGamma = 0.15;
SoCamera *Camera;
EmissiveSampleSet *eset;
SoTransform *cylXform;
SoTransform *cylXform2;

const int valSize=2;
const int maxDim = 256;
char dataFileName[] = "bunny2";
int dims[3] = { 64, 64, 46 };
double data[64*64*46];


unsigned int
readVal(unsigned char *ptr, int valSize, int bigEndian=0){
  int i;
  unsigned int c = 0;

  if (bigEndian){
    for (i=valSize-1; i>=0; i--){
      c = (c << 8) | ptr[i];
    }
  }
  else {
    for (i=0; i<valSize; i++){
      c = (c << 8) | ptr[i];
    }
  }

  // tweak for stanford bunny
  if (c == 0xf830)
    c = 0;
  
  return c;
}

void
readScalarField(){

  // open file in binary
 
  FILE *dataFile = fopen(dataFileName, "rb");
  if (!dataFile){
    cerr << "Could not open file: " << dataFileName << endl;
    exit(1);
  }
  
  int c;
  double d;
  double dmax = 0.0, dmin = HUGE_VAL;
  int numd = 0;
  double avgd = 0.0;
  unsigned char buffer[maxDim*maxDim*valSize];
  
  // iterate through each grid point
  for (int k = 0; k < dims[2]; k++) {
    
    // read in data
    cerr << "Reading in slice " << k+1 << endl;
    fread(buffer, valSize, dims[0]*dims[1], dataFile);
    
    for (int j = 0; j < dims[1]; j++) {
      for (int i = 0; i < dims[0]; i++) {

	// compute data
	c = readVal(&buffer[(j*dims[0]+i)*valSize], valSize, 0);
	d = (double)(c)/65536.0;
      
	// statistics 
	numd++;
	avgd+=d;
	if      (d>dmax) dmax = d;
	else if (d<dmin) dmin = d;
	
	data[k*dims[1]*dims[0] + j*dims[0]+i] = d;
	//data[index] = sin(p[0]*p[0]) + cos(p[1]*p[1]) + p[2]*p[2];

      }
    }
  }

  avgd/=numd;
  cerr << "Max: " << dmax << "\nMin: " << dmin
       << "\nAvg:" << avgd<< endl;
  fclose(dataFile);

}

////////////////////////////////////////////////////////////////////////
// Returns scalar value of field
////////////////////////////////////////////////////////////////////////

const double d = exp(2.0)-1.0;
const SbVec3f min = SbVec3f(-1.0, 1.0, -1.0);
const SbVec3f maxDiff = SbVec3f( 2.0, 2.0, 2.0);
const SbVec3f max = min + maxDiff;
//const SbVec3f d1 = SbVec3f(-0.5, -0.5, -0.5) + max;
//const SbVec3f d2 = SbVec3f(0.75, 0.75, 0.75) + min;
const SbVec3f d1 = SbVec3f( 0.5, 1.5, 0.0 );
const SbVec3f d2 = SbVec3f(-0.5, 1.5, 0.0 );

double
getFuncValue( SbVec3f p ){
  // Define function on y<3.0

  if (p[0]<min[0] || p[0]>max[0] ||
      p[1]<min[1] || p[1]>max[1] ||
      p[2]<min[2] || p[2]>max[2])
    return(0.0);

  p = p-min;
  p[0] = p[0]/2.0*(double)dims[0];
  p[1] = p[1]/2.0*(double)dims[1];
  p[2] = p[2]/2.0*(double)dims[2];

  return (data[ (int)(p[2])*dims[1]*dims[0] + (int)p[1]*dims[0] + (int)p[0] ]);
}

////////////////////////////////////////////////////////////////////////
// Returns gradient value of field * scalar value
////////////////////////////////////////////////////////////////////////

const double eps = 0.005;
const SbVec3f xeps = SbVec3f(eps, 0.0, 0.0);
const SbVec3f yeps = SbVec3f(0.0, eps, 0.0);
const SbVec3f zeps = SbVec3f(0.0, 0.0, eps);

void
getFuncValues( SbVec3f p, SbVec3f &grad, double &v ){
    v = getFuncValue(p);
    return;
    grad = SbVec3f((getFuncValue(p+xeps)-getFuncValue(p-xeps))/2.0*eps,
		   (getFuncValue(p+yeps)-getFuncValue(p-yeps))/2.0*eps,
		   (getFuncValue(p+zeps)-getFuncValue(p-zeps))/2.0*eps);
    return;
}

SoMaterial *esetMat;

SoSeparator *
makeScene (){

  readScalarField();

  SoSeparator *scene = new SoSeparator();
  eset = new EmissiveSampleSet();
  SoSeparator *esetSep = new SoSeparator();
  esetSep->addChild(eset);

  // Cube
  SoSeparator *cubeSep = new SoSeparator();
  SoMaterial *cubeMat = new SoMaterial();
  cubeMat->diffuseColor.setValue(1.0, 1.0, 1.0);
  cubeSep->addChild(cubeMat);
  
  /*     ____
       /|1  /|5    z
      / |  / |     |
     /__|0=min4    |__y
    2|  / |6=max  /
     | /  | /    x
    3|/___|7

  */
    
  static float vertices[8][3] =
  {  { min[0], min[1], min[2]},   // 0
     { min[0], min[1], max[2]},   // 1
     { max[0], min[1], max[2]},   // 2
     { max[0], min[1], min[2]},   // 3
     { min[0], max[1], min[2]},   // 4
     { min[0], max[1], max[2]},   // 5
     { max[0], max[1], max[2]},   // 6
     { max[0], max[1], min[2]}    // 7
  };

  static int32_t indices[40] =
  {
    0, 1, 2, 3,  -1,
    2, 6, 7, 3,  -1,
    4, 7, 6, 5,  -1,
    0, 4, 5, 1,  -1,
    1, 5, 6, 2,  -1,
    0, 3, 7, 4,  -1
  };

  // coordinates
  SoCoordinate3 *cubeCoords = new SoCoordinate3();
  cubeCoords->point.setValues(0,8, vertices);
  cubeSep->addChild(cubeCoords);

  // line set
  SoIndexedLineSet *cubeLineSet = new SoIndexedLineSet();
  cubeLineSet->coordIndex.setValues(0, 40, indices);
  cubeSep->addChild( cubeLineSet );

  // add to scene
  scene->addChild(cubeSep);
  scene->addChild(esetSep);
  
  return scene;
}

double globalrmax=0.0;

void
animate(void *, SoSensor *)
{
  int sample;
  double rmin=HUGE, rmax=-HUGE;
  
  // while rendMaxSamples not reached
  if (rendSamples<=rendSamplesMax){
  
    // shoot several rays
    for (sample=0; sample<=rendSamplesPerPass;){
  
      // shoot a ray
      SbVec3f p0, dp, intersection, minIs;
      dp.setValue(1.0, 1.0, 1.0);
      while (dp[1]<0.75 || dp.length()>1.0){
	dp = SbVec3f(drand48()*2.0-1.0,
		     drand48()*2.0-1.0,
		     drand48()*2.0-1.0);
      }
      dp.normalize();
      p0.setValue(0.0, 0.0, 0.0);

      // March through
      SbVec3f loc;
      SbVec3f g, x, y, z;
      double r = 0.0;
      double v;
      double dot;

      for (loc=p0+dp*1.0; loc[1]<3.0 && loc[0]<1.0 && loc[2]<1.0;
	   loc+=dp*0.00025){
	getFuncValues(loc, g, v);

	// Gradient calculation
	/*
	x = g;          // Surface Normal
	x.normalize();
	y = -dp;        // Vout
	dot = x.dot(y);
	*/
	
	// isolate v = c
	static const double c = 0.039;
	static const double w = 0.01;
	v = exp( -(v-c)*(v-c)/(w*w) );

	// Add
	if (dot<0.0)
	  dot = -dot;
	r += v; //*dot;

	if (r<rmin) rmin = r;
	else if (r>rmax) rmax = r;

      }

      //if (r==0.0)
      //  continue;

      //r/= (10.0/(rendSamples+1))*1000.0 +
      //(1.0-10.0/(rendSamples+1))*globalrmax;
      r/=2250.0;
      
      eset->location.set1Value(rendSamples, p0+1.0*dp);
      eset->normal.set1Value(rendSamples, dp);
      eset->color.set1Value(rendSamples, SbColor(r, r, r));
      eset->singleRadius.setValue(sqrt(5.76/(rendSamples+1)/3.14159)*1.0);

      rendSamples++;
      sample++;
    }
  }

  cerr << "Rmin: " << rmin << "   Rmax: " << rmax << endl;
  if (rmax>globalrmax)
    globalrmax = rmax;
  
  return;
}

////////////////////////////////////////////////////////////
// This code is from Dr. David C. Banks
//
// The main program gets a window from X, 
// puts an interactive 3D examinerViewer in it, 
// attaches the scene graph to the viewer,
// renders the scene graph,
// and hands control over to the Inventor main loop.

void
main(int argc, char* argv[])
{

  // read command line arguments
  int i;
  for (i=1; i<argc; i++){
    if (strcmp(argv[i], "-s") == 0){
      rendSamplesMax = atoi(argv[i+1]);
      /*
      if (rendSamplesMax>dims){
	numSamples = dim;
	cerr << "Too many samples. Using " << dim << ".\n";
      }
      */
    }
    else if (strcmp(argv[i], "-g") == 0)
      myGamma = atof(argv[i+1]);
    else if (strcmp(argv[i], "-h") == 0){
      cerr << "\nUsage: " << argv[0]
	   << " [ -s samples ] [ -g gamma ]\n\n";
      exit(1);
    }
  }

    
  // Initialize Inventor and Xt
  Widget myWindow = SoXt::init(argv[0]);

  // Initialize new classes
  EmissiveSampleSet::initClass();
  srand48(666);
  
  if (myWindow == NULL) exit(1);

  // SoGLRenderAction *rend = new SoGLRenderAction();
  //rend->setTransparencyType(SoGLRenderAction::SORTED_OBJECT_ADD);
  
  // Create a scene graph
  SoSeparator *root = makeScene();

  SoTimerSensor *sensor = new SoTimerSensor(animate, root);
  //  sensor->setInterval(SbTime(0.015)); // 70 frames per second
  sensor->setInterval(SbTime(0.015)); // 70 frames per second
  sensor->schedule();
  
  SoXtExaminerViewer *myViewer = new SoXtExaminerViewer(myWindow);
  myViewer->setSceneGraph(root);
  Camera = myViewer->getCamera();
  myViewer->setTitle("Hi mom");
  myViewer->setTransparencyType(SoGLRenderAction::SORTED_OBJECT_BLEND);
  //  myViewer->setBackgroundColor(SbColor(0.035, 0.035, 0.035));
  myViewer->setBackgroundColor(SbColor(0.035, 0.35, 0.035));
  myViewer->show();

  SoWriteAction wa;  // Write out the scene graph just for demo purposes 
  //wa.apply(root);    // The scenegraph can be displayed using
		     // ivview, SceneViewer, or gview .
		     // It can also be displayed in a Web browser by
		     // changing its header from
		     //
		     // #Inventor V2.1 ascii
		     // 
		     // to
		     //
		     // #VRML V1.0 ascii
		     //
		     // to exploit the (deprecated) original version
		     // of the Virtual Reality Modelling Language
		     // file format.
		     // 
		     // Visit http://www.vrml.org/ (deprecated URL) or
		     // http://www.web3d.org/ for info on 
		     // VRML begat Web3D begat X3D.


  SoXt::show(myWindow);
  SoXt::mainLoop();
  }
